@@ -1,5 +1,5 @@
 /*************************************************************************** 
- * RT2400/RT2500 SourceForge Project - http://rt2x00.serialmonkey.com      * 
+ * RT2x00 SourceForge Project - http://rt2x00.sourceforge.net              * 
  *                                                                         * 
  *   This program is free software; you can redistribute it and/or modify  * 
  *   it under the terms of the GNU General Public License as published by  * 
@@ -18,22 +18,26 @@
  *                                                                         * 
  *   Licensed under the GNU GPL                                            * 
  *   Original code supplied under license from RaLink Inc, 2004.           * 
- ***************************************************************************/ 
+ ***************************************************************************/
 
- /*************************************************************************** 
- *      Module Name: sanity.c 
- *              
- *      Abstract: 
- *              
- *      Revision History: 
- *      Who             When            What 
- *      --------        -----------     ----------------------------- 
- *      MarkW           8th  Dec 04     Baseline code  
- ***************************************************************************/ 
+/*************************************************************************** 
+ *	Module Name:	sanityc
+ *
+ *	Abstract:
+ *
+ *	Revision History:
+ *	Who		When		What
+ *	--------	----------	-------------------------------
+ *	Name		Date		Modification logs
+ *	Jan Lee		2005-06-01	Release
+ ***************************************************************************/
 
 #include "rt_config.h"
 
-UCHAR   WPA_OUI[] = {0x00, 0x50, 0xf2, 0x01};
+UCHAR	WPA_OUI[] = {0x00, 0x50, 0xf2, 0x01};
+UCHAR	RSN_OUI[] = {0x00, 0x0f, 0xac};
+
+extern 	UCHAR	BCAST[];
 
 /* 
     ==========================================================================
@@ -44,7 +48,7 @@ UCHAR   WPA_OUI[] = {0x00, 0x50, 0xf2, 0x01};
     ==========================================================================
  */
 BOOLEAN MlmeScanReqSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT UCHAR *BssType, 
@@ -53,13 +57,13 @@ BOOLEAN MlmeScanReqSanity(
     OUT UCHAR *ScanType) 
 {
     MLME_SCAN_REQ_STRUCT *Info;
-
+    
     Info = (MLME_SCAN_REQ_STRUCT *)(Msg);
     *BssType = Info->BssType;
     *SsidLen = Info->SsidLen;
     memcpy(Ssid, Info->Ssid, *SsidLen);
     *ScanType = Info->ScanType;
-
+    
     if ((*BssType == BSS_INFRA || *BssType == BSS_INDEP || *BssType == BSS_ANY) &&
        (*ScanType == SCAN_ACTIVE || *ScanType == SCAN_PASSIVE)) 
         return TRUE;
@@ -79,7 +83,7 @@ BOOLEAN MlmeScanReqSanity(
     ==========================================================================
  */
 BOOLEAN MlmeStartReqSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT CHAR Ssid[], 
@@ -107,10 +111,13 @@ BOOLEAN MlmeStartReqSanity(
         MLME message sanity check
     Return:
         TRUE if all parameters are OK, FALSE otherwise
+        
+	IRQL = DISPATCH_LEVEL
+
     ==========================================================================
  */
 BOOLEAN MlmeAssocReqSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT MACADDR *ApAddr, 
@@ -125,7 +132,7 @@ BOOLEAN MlmeAssocReqSanity(
     COPY_MAC_ADDR(ApAddr, &Info->Addr);                   // AP address
     *CapabilityInfo = Info->CapabilityInfo;               // capability info
     *ListenIntv = Info->ListenIntv;
-
+    
     return TRUE;
 }
 
@@ -138,7 +145,7 @@ BOOLEAN MlmeAssocReqSanity(
     ==========================================================================
  */
 BOOLEAN MlmeAuthReqSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT MACADDR *Addr, 
@@ -151,7 +158,7 @@ BOOLEAN MlmeAuthReqSanity(
     COPY_MAC_ADDR(Addr, &Info->Addr);
     *Timeout = Info->Timeout;
     *Alg = Info->Alg;
-
+    
     if ((*Alg == Ndis802_11AuthModeShared || *Alg == Ndis802_11AuthModeOpen) && !MAC_ADDR_IS_GROUP(*Addr)) 
     {
         return TRUE;
@@ -169,10 +176,13 @@ BOOLEAN MlmeAuthReqSanity(
         MLME message sanity check
     Return:
         TRUE if all parameters are OK, FALSE otherwise
+        
+	IRQL = DISPATCH_LEVEL
+
     ==========================================================================
  */
 BOOLEAN PeerAssocRspSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT MACADDR *Addr2, 
@@ -185,21 +195,21 @@ BOOLEAN PeerAssocRspSanity(
 {
     CHAR          IeType, *Ptr;
     MACFRAME     *Fr = (MACFRAME *)Msg;
-    PBEACON_EID_STRUCT  eid_ptr;
-
+    PBEACON_EID_STRUCT	eid_ptr;
+    
     COPY_MAC_ADDR(Addr2, &Fr->Hdr.Addr2);
     Ptr = Fr->Octet;
-
+        
     memcpy(CapabilityInfo, &Fr->Octet[0], 2);
     memcpy(Status,         &Fr->Octet[2], 2);
-    // Mask out unnecessary capability information
-    *CapabilityInfo &= SUPPORTED_CAPABILITY_INFO;
-
+	// Mask out unnecessary capability information
+	*CapabilityInfo &= SUPPORTED_CAPABILITY_INFO;
+    
     if (*Status == MLME_SUCCESS) 
     {
         memcpy(Aid, &Fr->Octet[4], 2);
         *Aid = (*Aid) & 0x3fff; // AID is low 14-bit
-
+        
         // -- get supported rates from payload and advance the pointer
         IeType = Fr->Octet[6];
         *RatesLen = Fr->Octet[7];
@@ -209,13 +219,13 @@ BOOLEAN PeerAssocRspSanity(
             return FALSE;
         } 
         else 
-            memcpy(Rates, &Fr->Octet[8], *RatesLen);
+       		memcpy(Rates, &Fr->Octet[8], *RatesLen);
 
         // many AP implement proprietary IEs in non-standard order, we'd better
         // tolerate mis-ordered IEs to get best compatibility
         *ExtendedRateIeExist = FALSE;
         eid_ptr = (PBEACON_EID_STRUCT) &Fr->Octet[8 + (*RatesLen)];
-
+            
         // get variable fields from payload and advance the pointer
         while (((UCHAR*)eid_ptr + eid_ptr->Len + 1) < ((UCHAR*)Fr + MsgLen))
         {
@@ -238,12 +248,11 @@ BOOLEAN PeerAssocRspSanity(
                     DBGPRINT(RT_DEBUG_TRACE, "PeerAssocRspSanity - ignore unrecognized EID = %d\n", eid_ptr->Eid);
                     break;
             }
-
-            eid_ptr = (PBEACON_EID_STRUCT)((UCHAR*)eid_ptr + 2 + eid_ptr->Len);
+                   
+            eid_ptr = (PBEACON_EID_STRUCT)((UCHAR*)eid_ptr + 2 + eid_ptr->Len);        
         }
-
     }
-
+    
     return TRUE;
 }
 
@@ -253,10 +262,13 @@ BOOLEAN PeerAssocRspSanity(
         MLME message sanity check
     Return:
         TRUE if all parameters are OK, FALSE otherwise
+        
+	IRQL = DISPATCH_LEVEL
+
     ==========================================================================
  */
 BOOLEAN PeerDisassocSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT MACADDR *Addr2, 
@@ -276,10 +288,13 @@ BOOLEAN PeerDisassocSanity(
         MLME message sanity check
     Return:
         TRUE if all parameters are OK, FALSE otherwise
+        
+	IRQL = DISPATCH_LEVEL
+
     ==========================================================================
  */
 BOOLEAN PeerDeauthSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT MACADDR *Addr2, 
@@ -299,10 +314,13 @@ BOOLEAN PeerDeauthSanity(
         MLME message sanity check
     Return:
         TRUE if all parameters are OK, FALSE otherwise
+        
+	IRQL = DISPATCH_LEVEL
+
     ==========================================================================
  */
 BOOLEAN PeerAuthSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT MACADDR *Addr, 
@@ -360,10 +378,13 @@ BOOLEAN PeerAuthSanity(
         MLME message sanity check
     Return:
         TRUE if all parameters are OK, FALSE otherwise
+        
+	IRQL = DISPATCH_LEVEL
+
     ==========================================================================
  */
 BOOLEAN PeerProbeReqSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT MACADDR *Addr2,
@@ -373,8 +394,8 @@ BOOLEAN PeerProbeReqSanity(
 //    OUT UCHAR *RatesLen) 
 {
     UCHAR Idx;
-    UCHAR	RateLen;
-    CHAR          IeType;
+	UCHAR	RateLen;
+	CHAR          IeType;
     MACFRAME *Fr = (MACFRAME *)Msg;
 
     COPY_MAC_ADDR(Addr2, &Fr->Hdr.Addr2);
@@ -388,23 +409,22 @@ BOOLEAN PeerProbeReqSanity(
     *SsidLen = Fr->Octet[1];
     memcpy(Ssid, &Fr->Octet[2], *SsidLen);
 
-#if 1    
     Idx = *SsidLen + 2;
 
-    // -- get supported rates from payload and advance the pointer
-    IeType = Fr->Octet[Idx];
-    RateLen = Fr->Octet[Idx + 1];
-    if (IeType != IE_SUPP_RATES) 
+	// -- get supported rates from payload and advance the pointer
+	IeType = Fr->Octet[Idx];
+	RateLen = Fr->Octet[Idx + 1];
+	if (IeType != IE_SUPP_RATES) 
     {
         DBGPRINT(RT_DEBUG_TRACE, "PeerProbeReqSanity fail - wrong SupportRates IE(Type=%d,Len=%d)\n",Fr->Octet[Idx],Fr->Octet[Idx+1]);
         return FALSE;
     }
-    else 
-    {
-        if ((pAd->PortCfg.AdhocMode == 2) && (RateLen < 8))
-            return (FALSE);
-    }
-#endif
+	else
+	{
+		if ((pAd->PortCfg.AdhocMode == 2) && (RateLen < 8))
+			return (FALSE);
+	}
+
     return TRUE;
 }
 
@@ -414,10 +434,13 @@ BOOLEAN PeerProbeReqSanity(
         MLME message sanity check
     Return:
         TRUE if all parameters are OK, FALSE otherwise
+        
+	IRQL = DISPATCH_LEVEL
+
     ==========================================================================
  */
 BOOLEAN PeerBeaconAndProbeRspSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN VOID *Msg, 
     IN ULONG MsgLen, 
     OUT MACADDR *Addr2, 
@@ -445,31 +468,33 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
 	OUT UCHAR *SupRateLen,
 	OUT UCHAR ExtRate[],
 	OUT UCHAR *ExtRateLen,
+    OUT UCHAR *LengthVIE,
     OUT	PNDIS_802_11_VARIABLE_IEs pVIE) 
 {
-    CHAR                *Ptr, TimLen;
-    MACFRAME            *Fr;
-    PBEACON_EID_STRUCT  eid_ptr;
-    UCHAR               SubType;
-    UCHAR               Sanity;
+    CHAR				*Ptr, TimLen;
+    MACFRAME			*Fr;
+    PBEACON_EID_STRUCT	eid_ptr;
+    UCHAR				SubType;
+	UCHAR				Sanity;
 
-    // Add for 3 necessary EID field check
-    Sanity = 0;
+	// Add for 3 necessary EID field check
+	Sanity = 0;
 
     *ExtendedRateIeExist = FALSE;
     *Erp = 0;
-
+	*LengthVIE = 0;					// Set the length of VIE to init value 0
+    
     Fr = (MACFRAME *)Msg;
-
+    
     // get subtype from header
     SubType = (UCHAR)Fr->Hdr.SubType;
 
     // get Addr2 and BSSID from header
     COPY_MAC_ADDR(Addr2, &Fr->Hdr.Addr2);
     COPY_MAC_ADDR(Bssid, &Fr->Hdr.Addr3);
-
+    
     Ptr = Fr->Octet;
-
+    
     // get timestamp from payload and advance the pointer
     memcpy(Timestamp, Ptr, TIMESTAMP_LEN);
     Ptr += TIMESTAMP_LEN;
@@ -490,9 +515,9 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
         *BssType = BSS_INDEP;
     }
 
-    // Mask out unnecessary capability information
-    *CapabilityInfo &= SUPPORTED_CAPABILITY_INFO;
-    
+	// Mask out unnecessary capability information
+	*CapabilityInfo &= SUPPORTED_CAPABILITY_INFO;
+	
     eid_ptr = (PBEACON_EID_STRUCT) Ptr;
 
     // get variable fields from payload and advance the pointer
@@ -501,16 +526,14 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
         switch(eid_ptr->Eid)
         {
             case IE_SSID:
-                // Already has one SSID EID in this beacon, ignore the second one
+				// Already has one SSID EID in this beacon, ignore the second one
 				if (Sanity & 0x1)
-					break;
+					break;				
                 if(eid_ptr->Len <= MAX_LEN_OF_SSID)
                 {
                     memcpy(Ssid, eid_ptr->Octet, eid_ptr->Len);
-		    memset(Ssid + eid_ptr->Len,0,1);
                     *SsidLen = eid_ptr->Len;
-                    Sanity |= 0x1;
-                    //DBGPRINT(RT_DEBUG_TRACE, "PeerBeaconAndProbeRspSanity - ESSID=%s Len=%d\n",Ssid,eid_ptr->Len);
+					Sanity |= 0x1;
                 }
                 else
                 {
@@ -538,11 +561,12 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
             		}
             		*RateLen = i;
 					Sanity |= 0x2;
-
-                    // Copy supported rate from desired AP's beacon. We are trying to match
+					// Copy supported rate from desired AP's beacon. We are trying to match
 					// AP's supported and extended rate settings.
 					memcpy(SupRate, eid_ptr->Octet, eid_ptr->Len);
 					*SupRateLen = eid_ptr->Len;
+					// Check against the supported rates
+					RTMPCheckRates(pAd, SupRate, SupRateLen);
                 }
                 else
                 {
@@ -558,13 +582,13 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
             case IE_DS_PARM:
                 if(eid_ptr->Len == 1)
                 {
-                    *Channel = *eid_ptr->Octet;
+                    *Channel = *eid_ptr->Octet;                    
                     if (ChannelSanity(pAd, *Channel) == 0)
                     {
                         DBGPRINT(RT_DEBUG_TRACE, "PeerBeaconAndProbeRspSanity - wrong IE_DS_PARM (ch=%d)\n",*Channel);
                         return FALSE;
                     }
-                    Sanity |= 0x4;
+					Sanity |= 0x4;
                 }
                 else
                 {
@@ -605,22 +629,22 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
                 }
                 break;
 
-            // New for WPA
-            case IE_WPA:
-                // Check the OUI version, filter out non-standard usage
-                if (RTMPEqualMemory(eid_ptr->Octet, WPA_OUI, 4))
-                {
-                    // Copy to pVIE which will report to microsoft bssid list.
-                    pVIE->ElementID = eid_ptr->Eid;
-                    pVIE->Length = eid_ptr->Len;
-                    memcpy(pVIE->data, eid_ptr->Octet, eid_ptr->Len);
-                }
-                DBGPRINT(RT_DEBUG_INFO, "PeerBeaconAndProbeRspSanity - Receive IE_WPA\n");
-                break;
+			// New for WPA
+			case IE_WPA:
+				// Check the OUI version, filter out non-standard usage
+				if ((memcmp(eid_ptr->Octet, WPA_OUI, 4) == 0))
+				{
+					// Copy to pVIE which will report to microsoft bssid list.
+					Ptr = (PUCHAR) pVIE;
+					memcpy(Ptr + *LengthVIE, &eid_ptr->Eid, eid_ptr->Len + 2);
+					*LengthVIE += (eid_ptr->Len + 2);
+				}
+				DBGPRINT(RT_DEBUG_INFO, "PeerBeaconAndProbeRspSanity - Receive IE_WPA\n");
+				break;
 
-            case IE_EXT_SUPP_RATES:
-                // concatenate all extended rates to Rates[] and RateLen
-                *ExtendedRateIeExist = TRUE;
+			case IE_EXT_SUPP_RATES:
+			    // concatenate all extended rates to Rates[] and RateLen
+			    *ExtendedRateIeExist = TRUE;
                 if (eid_ptr->Len <= MAX_LEN_OF_SUPPORTED_RATES)
                 {
             		int		index;
@@ -641,12 +665,14 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
             		        break;
             		}
             		*RateLen = i;
-                    // Copy extended rate from desired AP's beacon. We are trying to match
+					// Copy extended rate from desired AP's beacon. We are trying to match
 					// AP's supported and extended rate settings.
 					memcpy(ExtRate, eid_ptr->Octet, eid_ptr->Len);
 					*ExtRateLen = eid_ptr->Len;
-                }
-                break;
+					// Check against the supported rates
+					RTMPCheckRates(pAd, ExtRate, ExtRateLen);
+				}
+			    break;
 
             case IE_ERP:
                 if (eid_ptr->Len == 1)
@@ -655,15 +681,27 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
                 }
                 break;
                 
+			// WPA2 & 802.11i RSN
+			case IE_RSN:
+				// There is no OUI for version anymore, check the group cipher OUI before copying
+				if ((memcmp(eid_ptr->Octet + 2, RSN_OUI, 3) == 0))
+				{
+					// Copy to pVIE which will report to microsoft bssid list.
+					Ptr = (PUCHAR) pVIE;
+					memcpy(Ptr + *LengthVIE, &eid_ptr->Eid, eid_ptr->Len + 2);
+					*LengthVIE += (eid_ptr->Len + 2);
+				}
+                DBGPRINT(RT_DEBUG_INFO, "IE_RSN length = %d\n", eid_ptr->Len);
+                break;
+				
             default:
                 DBGPRINT(RT_DEBUG_INFO, "PeerBeaconAndProbeRspSanity - unrecognized EID = %d\n", eid_ptr->Eid);
                 break;
         }
         
-        eid_ptr = (PBEACON_EID_STRUCT)((UCHAR*)eid_ptr + 2 + eid_ptr->Len);
+        eid_ptr = (PBEACON_EID_STRUCT)((UCHAR*)eid_ptr + 2 + eid_ptr->Len);        
     }
 
-
     // in 802.11a band, AP may skip this DS IE in their BEACON
     if ((pAd->PortCfg.Channel > 14) && ((Sanity & 0x04)==0))
     {
@@ -686,6 +724,9 @@ BOOLEAN PeerBeaconAndProbeRspSanity(
 /* 
     ==========================================================================
     Description:
+        
+	IRQL = DISPATCH_LEVEL
+
     ==========================================================================
  */
 BOOLEAN GetTimBit(
@@ -701,18 +742,18 @@ BOOLEAN GetTimBit(
     CHAR          *IdxPtr;
 
     IdxPtr = Ptr;
-
+    
     IdxPtr ++;
     *TimLen = *IdxPtr;
-
+    
     // get DTIM Count from TIM element
     IdxPtr ++;
     *DtimCount = *IdxPtr;
-
+    
     // get DTIM Period from TIM element
     IdxPtr++;
     *DtimPeriod = *IdxPtr;
-
+        
     // get Bitmap Control from TIM element
     IdxPtr++;
     BitCntl = *IdxPtr;
@@ -722,11 +763,10 @@ BOOLEAN GetTimBit(
     else 
         *BcastFlag = FALSE;
     
-#if 1
     // Parse Partial Virtual Bitmap from TIM element
     N1 = BitCntl & 0xfe;    // N1 is the first bitmap byte#
     N2 = *TimLen - 4 + N1;  // N2 is the last bitmap byte#
-
+    
     if ((Aid < (N1 << 3)) || (Aid >= ((N2 + 1) << 3)))
         *MessageToMe = FALSE;
     else
@@ -744,14 +784,10 @@ BOOLEAN GetTimBit(
         else 
             *MessageToMe = FALSE;
     }
-#else
-    *MessageToMe = FALSE;
-#endif
 
     return TRUE;
 }
 
-
 /*!
  *  \brief Get legacy bit, right now for 11b it is always 0
  *  \param
@@ -767,8 +803,9 @@ BOOLEAN GetLegacy(
     return TRUE;
 }
 
+// IRQL = DISPATCH_LEVEL
 UCHAR ChannelSanity(
-    IN PRTMP_ADAPTER pAd, 
+    IN PRT2570ADAPTER pAd, 
     IN UCHAR channel)
 {
     UCHAR index;
@@ -779,53 +816,139 @@ UCHAR ChannelSanity(
             return 1;
     }
     return 0;
-
-#if 0    
-    switch (pAd->PortCfg.CountryRegion)
-    {
-        case REGION_FCC:    // 1 - 11
-            if ((channel > 0) && (channel < 12))
-                return 1;
-            break;
-            
-        case REGION_IC:     // 1 -11
-            if ((channel > 0) && (channel < 12))
-                return 1;
-            break;
-            
-        case REGION_ETSI:   // 1 - 13
-            if ((channel > 0) && (channel < 14))
-                return 1;
-            break;
-            
-        case REGION_SPAIN:  // 10 - 11
-            if ((channel > 9) && (channel < 12))
-                return 1;
-            break;
-            
-        case REGION_FRANCE: // 10 -13
-            if ((channel > 9) && (channel < 14))
-                return 1;
-            break;
-            
-        case REGION_MKK:    // 14
-            if (channel == 14)              
-                return 1;
-            break;
-            
-        case REGION_MKK1:   // 1 - 14
-            if ((channel > 0) && (channel < 15))
-                return 1;
-            break;
-            
-        case REGION_ISRAEL: // 3 - 9
-            if ((channel > 2) && (channel < 10))
-                return 1;
-            break;
-            
-        default:            // Error
-            return 0;           
-    }   
-    return (0);
-#endif    
 }
+
+/*
+	========================================================================
+	Routine Description:
+		Sanity check NetworkType (11b, 11g or 11a)
+		
+	Arguments:
+		pBss - Pointer to BSS table.
+
+	Return Value:
+        Ndis802_11DS .......(11b)
+        Ndis802_11OFDM24....(11g)
+        Ndis802_11OFDM5.....(11a)
+        
+	IRQL = DISPATCH_LEVEL
+	
+	========================================================================
+*/
+NDIS_802_11_NETWORK_TYPE NetworkTypeInUseSanity(
+    IN PBSS_ENTRY pBss)
+{
+	NDIS_802_11_NETWORK_TYPE	NetWorkType = Ndis802_11DS;
+	UCHAR						rate, i;
+
+	if (pBss->Channel <= 14)
+	{
+		for (i = 0; i < pBss->RatesLen; i++)
+		{
+			rate = pBss->Rates[i] & 0x7f; // Mask out basic rate set bit		
+			if ((rate == 2) || (rate == 4) || (rate == 11) || (rate == 22))
+			{
+				//
+				// basic rate means Ndis802_11DS
+				//
+				NetWorkType = Ndis802_11DS;
+				continue;
+			}
+			else
+			{
+				//
+				// Otherwise (even rate > 108) means Ndis802_11OFDM24
+				//
+				NetWorkType = Ndis802_11OFDM24;
+				break;
+			}	
+		}
+	}
+	else
+	{
+		NetWorkType = Ndis802_11OFDM5;
+	}
+
+	return NetWorkType;
+}
+
+/*
+	========================================================================
+	Routine Description:
+		Sanity check pairwise key on pAdapter->PortCfg.AuthMode < Ndis802_11AuthModeWPA
+		
+	Arguments:
+		pAdapter - Pointer to our adapter
+		pBuf 	 - Pointer to NDIS_802_11_KEY structure
+
+	Return Value:
+		NDIS_STATUS_SUCCESS
+		NDIS_STATUS_INVALID_DATA
+        
+	IRQL = DISPATCH_LEVEL
+
+	Note:
+		For OID_802_11_ADD_KEY setting, on old wep stuff also need to verify
+		the structure of NIDS_802_11_KEY
+	========================================================================
+*/
+NDIS_STATUS	RTMPWPAWepKeySanity(
+	IN	PRT2570ADAPTER	pAdapter,
+	IN	PVOID			pBuf)
+{
+	PNDIS_802_11_KEY	pKey;
+	ULONG				KeyIdx;
+	NDIS_STATUS			Status = NDIS_STATUS_SUCCESS;
+	BOOLEAN 			bTxKey; 		// Set the key as transmit key
+	BOOLEAN 			bPairwise;		// Indicate the key is pairwise key
+
+	pKey = (PNDIS_802_11_KEY) pBuf;
+	KeyIdx = pKey->KeyIndex & 0x0fffffff;
+	// Bit 31 of Add-key, Tx Key
+	bTxKey		   = (pKey->KeyIndex & 0x80000000) ? TRUE : FALSE;
+	// Bit 30 of Add-key PairwiseKey
+	bPairwise	   = (pKey->KeyIndex & 0x40000000) ? TRUE : FALSE;
+
+	// 1. Check Group / Pairwise Key
+	if (bPairwise)	// Pairwise Key
+	{
+		// 1. Check KeyIdx 
+		// it is a shared key
+		if (KeyIdx > 4)
+			return (-EINVAL);
+
+		// 2. Check bTx, it must be true, otherwise, return NDIS_STATUS_INVALID_DATA
+		if (bTxKey == FALSE)
+			return (-EINVAL);
+
+		// 3. If BSSID is not all 0xff, return NDIS_STATUS_INVALID_DATA
+		if ((memcmp(pKey->BSSID, BCAST, 6) == 0))
+			return (-EINVAL);
+
+		// check key length
+		if ((pKey->KeyLength != 5) && (pKey->KeyLength != 13))
+			return (-EINVAL);
+	}
+	else
+	{
+		// Group Key
+		// 1. Check BSSID, if not current BSSID or Bcast, return NDIS_STATUS_INVALID_DATA
+		if ((memcmp(&pKey->BSSID, &BCAST, 6) != 0) &&
+			(memcmp(&pKey->BSSID, &pAdapter->PortCfg.Bssid, 6) != 0))
+			return (-EINVAL);
+		// 2. Check Key index for supported Group Key
+		if (KeyIdx > 4)
+			return (-EINVAL);
+	}
+
+	pAdapter->PortCfg.SharedKey[KeyIdx].KeyLen = (UCHAR) pKey->KeyLength;
+	memcpy(pAdapter->PortCfg.SharedKey[KeyIdx].Key, &pKey->KeyMaterial, pKey->KeyLength);
+	if (pKey->KeyIndex & 0x80000000)
+	{
+		// Default key for tx (shared key)
+		pAdapter->PortCfg.DefaultKeyId = (UCHAR) KeyIdx;
+	}	
+
+	return (Status);
+}
+
